/* $Header: /fridge/cvs/libjcgi/cgi-ip.c,v 1.1 2001/12/10 22:56:34 justins Exp $ */
/*
   
   JCGI - cgi-ip.c            Copyright(c) 2001 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Utilities for manipulating IP numbers
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <assert.h>
#include <libjcgi.h>


static inline bool _process_ip(char *buf, unsigned int *ip) {
/* _process_ip
   Attempt to interpret a dotted-quad form IP address.  On success
   the number is stored to *ip, on failure false is returned.  The
   buffer will be modified.  */

   int succ;
   unsigned int ip1;
   unsigned int ip2;
   unsigned int ip3;
   unsigned int ip4;
   char *pip1;
   char *pip2;
   char *pip3;
   char *pip4;

   /* Find and delimit the separators */
   pip1 = buf;
   pip2 = pip1;
   while(*pip2 != '\0' && *pip2 != '.') ++pip2;
   if(*pip2 == '\0') return(false);
   *(pip2++) = '\0';
   pip3 = pip2;
   while(*pip3 != '\0' && *pip3 != '.') ++pip3;
   if(*pip3 == '\0') return(false);
   *(pip3++) = '\0';
   pip4 = pip3;
   while(*pip4 != '\0' && *pip4 != '.') ++pip4;
   if(*pip4 == '\0') return(false);
   *(pip4++) = '\0';
   
   /* Attempt to read each value */
   ip1 = strtoint(pip1, &succ);
   if(succ != 0) return(false);
   ip2 = strtoint(pip2, &succ);
   if(succ != 0) return(false);
   ip3 = strtoint(pip3, &succ);
   if(succ != 0) return(false);
   ip4 = strtoint(pip4, &succ);
   if(succ != 0) return(false);
   
   /* Construct the final IP */
   *ip  = (ip1 & 0xff) << 24;
   *ip |= (ip2 & 0xff) << 16;
   *ip |= (ip3 & 0xff) << 8;   
   *ip |= (ip4 & 0xff);
   return(true);

}


static inline bool _process_ip_mask(char *buf, unsigned int *ip) {
/* _process_ip_mask
   Similar to above, but we assume this is a mask.  So if we cannot parse
   the buffer as a single integer, try using a dotted quad form instead. 
   The buffer will be modified. */

   int succ;
   unsigned int mask;
   
   /* Attempt to parse as a single-value mask */
   mask = strtoint(buf, &succ);
   if(succ != 0) return(_process_ip(buf, ip));
   
   /* Success!  Interpret the mask */
   if(mask >= 32) {
      *ip = 0xffffffff;
      return(true);
   }
   if(mask <= 0) {
      *ip = 0x00000000;
      return(true);
   }
   *ip = 0xffffffff << (32 - mask);
   return(true);
   
}
   

bool ip_number_init(ip_number *ipn, const char *ip) {
/* ip_number_new
   Parses a new IP number.  IP may include a mask.  Returns true
   on success, or false if the IP string could not be parsed. */

   char buf[0x1000];
   unsigned int ipnum;
   unsigned int ipmask;
   char *p;
   
   /* Make sure we have a string */
   if(ip == NULL || ipn == NULL) return(false);
   strcopyb(buf, ip, sizeof(buf));
   
   /* Search for a mask delimiter */
   p = buf;
   while(*p != '\0' && *p != '/') ++p;
   if(*p == '/') *(p++) = '\0';
   
   /* Process the main IP number - must be a dotted quad! */
   if(!_process_ip(buf, &ipnum)) return(false);
   
   /* Process the mask; may be missing, 1 int, or dotted quad */
   ipmask = 0xffffffff;
   if(*p != '\0' && !_process_ip_mask(p, &ipmask)) return(false);
   
   /* Assign to the IP structure */
   ipn->ip_number = ipnum;
   ipn->ip_mask   = ipmask;
   return(true);

}


ip_number *ip_number_new(const char *ip) {
/* ip_number_new
   Constructs a new IP number.  IP may include a mask.  Returns a new
   structure on success, or NULL if alloc error occurred or the IP
   string could not be parsed. */

   ip_number *ipn;

   ipn = (ip_number *)malloc(sizeof(ip_number));
   if(!ip_number_init(ipn, ip)) {
      free(ipn);
      return(NULL);
   }
   return(ipn);

}


bool ip_number_match(const ip_number *ipmask, const ip_number *iptest) {
/* Returns true if the IP encoded in iptest is equivalent to the IP
   given in ipmask, subject to the mask defined in ipmask (so if the
   ipmask is 131.215.0.0/16, iptest = 131.215.5.10 will succeed).  
   Note that the iptest mask is IGNORED for this computation.  */

   unsigned int ipt;
   unsigned int ipm;

   if(ipmask == NULL || iptest == NULL) return(false);

   ipt  = iptest->ip_number;
   ipt &= ipmask->ip_mask;
   ipm  = ipmask->ip_number;
   ipm &= ipmask->ip_mask;
   return(ipt == ipm);

}


void ip_number_free(ip_number *ipn) {
/* Releases an IP structure */

   if(ipn == NULL) return;
   free(ipn);
   
}


void ip_number_release(ip_number **ipn) {
/* Releases an IP structure */
   
   if(ipn == NULL || *ipn == NULL) return;
   free(*ipn);
   *ipn = NULL;

}

